package ch.zhaw.pm2.multichat.client;

import ch.zhaw.pm2.multichat.protocol.ChatProtocolException;
import ch.zhaw.pm2.multichat.protocol.ConnectionHandler;
import ch.zhaw.pm2.multichat.protocol.NetworkHandler;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.beans.property.SimpleStringProperty;

import java.io.EOFException;
import java.io.IOException;
import java.net.SocketException;
import java.util.Scanner;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import static ch.zhaw.pm2.multichat.protocol.ConnectionHandler.State.*;

public class ClientConnectionHandler extends ConnectionHandler implements Runnable {

    private final Pattern messagePattern = Pattern.compile( "^(?:@(\\S*))?\\s*(.*)$" );

    private SimpleStringProperty userName;
    private SimpleObjectProperty<State> state;
    private ClientMessageList messages;
    private SimpleStringProperty serverAddress;
    private SimpleIntegerProperty serverPort;

    public ClientConnectionHandler(ClientMessageList messages)  {
        super();
        this.messages = messages;
        state = new SimpleObjectProperty<>(State.NEW);
        serverAddress = new SimpleStringProperty(NetworkHandler.DEFAULT_ADDRESS.getCanonicalHostName());
        serverPort = new SimpleIntegerProperty(NetworkHandler.DEFAULT_PORT);
        this.userName = new SimpleStringProperty(null);
    }

    public void initialize(String serverAddress, int serverPort, String userName) throws IOException {
        state.set(NEW);
        this.serverAddress.set(serverAddress);
        this.serverPort.set(serverPort);
        setConnection(NetworkHandler.openConnection(serverAddress, serverPort));
        this.userName.set((userName == null || userName.isBlank())? USER_NONE : userName);
    }

    public SimpleStringProperty getServerAddressProperty() { return serverAddress; }

    public SimpleIntegerProperty getServerPortProperty() { return serverPort; }

    public SimpleObjectProperty<State> getStateProperty() {
        return this.state;
    }

    public SimpleStringProperty getUserNameProperty() { return userName; }

    public void setState (State newState) {
        state.set(newState);
    }

    public void run () {
        startReceiving();
    }

    private void startReceiving() {
        System.out.println("Starting Connection Handler");
        try {
            System.out.println("Start receiving data...");
            while (getConnection().isAvailable()) {
                String data = getConnection().receive();
                processData(data);
            }
            System.out.println("Stopped recieving data");
        } catch (SocketException e) {
            System.out.println("Connection terminated locally");
            this.setState(DISCONNECTED);
            System.err.println("Unregistered because connection terminated" + e.getMessage());
        } catch (EOFException e) {
            System.out.println("Connection terminated by remote");
            this.setState(DISCONNECTED);
            System.err.println("Unregistered because connection terminated" + e.getMessage());
        } catch(IOException e) {
            System.err.println("Communication error" + e);
        } catch(ClassNotFoundException e) {
            System.err.println("Received object of unknown type" + e.getMessage());
        }
        System.out.println("Stopped Connection Handler");
    }

    public void stopReceiving() {
        System.out.println("Closing Connection Handler to Server");
        try {
            System.out.println("Stop receiving data...");
            getConnection().close();
            System.out.println("Stopped receiving data.");
        } catch (IOException e) {
            System.err.println("Failed to close connection." + e.getMessage());
        }
        System.out.println("Closed Connection Handler to Server");
    }

    private void processData(String data) {
        try {
            Scanner scanner = new Scanner(data);
            StringBuilder sender = new StringBuilder();
            StringBuilder reciever = new StringBuilder();
            StringBuilder type = new StringBuilder();
            StringBuilder payload = new StringBuilder();
            super.processData(scanner,sender,reciever,type,payload);

            // dispatch operation based on type parameter
            if (type.toString().equals(getDataTypeConnect())) {
                System.err.println("Illegal connect request from server");
            } else if (type.toString().equals(getDataTypeConfirm())) {
                caseConfirm(sender.toString(), reciever.toString(), payload.toString());
            } else if (type.toString().equals(getDataTypeDisconnect())) {
                caseDisconnect(sender.toString(),reciever.toString(),payload.toString());
            } else if (type.toString().equals(getDataTypeMessage())) {
                caseMessage(sender.toString(),reciever.toString(),payload.toString());
            } else if (type.toString().equals(getDataTypeError())) {
                caseError(sender.toString(), reciever.toString(), payload.toString());
            } else {
                System.out.println("Unknown data type received: " + type);
            }
        } catch (ChatProtocolException e) {
            System.err.println("Error while processing data: " + e.getMessage());
            sendData(USER_NONE, userName.get(), getDataTypeError(), e.getMessage());
        }
    }

    private void caseConfirm(String sender, String reciever, String payload) {
        if (state.get() == CONFIRM_CONNECT) {
            this.userName.set(reciever);
            this.serverPort.set(getConnection().getRemotePort());
            this.serverAddress.set(getConnection().getRemoteHost());
            messages.addMessage(new Message(Message.MessageType.INFO,sender,reciever,payload));
            System.out.println("CONFIRM: " + payload);
            this.setState(CONNECTED);
        } else if (state.get() == CONFIRM_DISCONNECT) {
            messages.addMessage(new Message(Message.MessageType.INFO,sender,reciever,payload));
            System.out.println("CONFIRM: " + payload);
            this.setState(DISCONNECTED);
        } else {
            System.err.println("Got unexpected confirm message: " + payload);
        }
    }

    private void caseDisconnect(String sender, String reciever, String payload) {
        if (state.get() == DISCONNECTED) {
            System.out.println("DISCONNECT: Already in disconnected: " + payload);
            return;
        }
        messages.addMessage(new Message(Message.MessageType.INFO,sender,reciever,payload));
        System.out.println("DISCONNECT: " + payload);
        this.setState(DISCONNECTED);
    }

    private void caseMessage(String sender, String reciever, String payload) {
        if (state.get() != CONNECTED) {
            System.out.println("MESSAGE: Illegal state " + state + " for message: " + payload);
            return;
        }
        messages.addMessage(new Message(Message.MessageType.MESSAGE,sender,reciever,payload));
        System.out.println("MESSAGE: From " + sender + " to " + reciever + ": "+  payload);
    }

    private void caseError(String sender, String reciever, String payload) {
        messages.addMessage(new Message(Message.MessageType.ERROR,sender,reciever,payload));
        System.out.println("ERROR: " + payload);
    }

    public void connect() throws ChatProtocolException {
        if (state.get() != NEW) throw new ChatProtocolException("Illegal state for connect: " + state);
        this.sendData(userName.get(), USER_NONE, getDataTypeConnect(),null);
        this.setState(CONFIRM_CONNECT);
    }

    public void disconnect() throws ChatProtocolException {
        if (state.get() != NEW && state.get() != CONNECTED) throw new ChatProtocolException("Illegal state for disconnect: " + state);
        this.sendData(userName.get(), USER_NONE, getDataTypeDisconnect(),null);
        this.setState(CONFIRM_DISCONNECT);
    }

    public boolean message(String messageString) throws ChatProtocolException {
        if (state.get() != CONNECTED) throw new ChatProtocolException("Illegal state for message: " + state);

        Matcher matcher = messagePattern.matcher(messageString);
        if (matcher.find()) {
            String receiver = matcher.group(1);
            String message = matcher.group(2);
            if(message.length() < 1){
                return false;
            }
            if (receiver == null || receiver.isBlank()) receiver = ClientConnectionHandler.USER_ALL;
            this.sendData(userName.get(), receiver, getDataTypeMessage(),message);
            return true;
        } else {
            return false;
        }
    }
}
