package ch.zhaw.pm2.multichat.server;

import ch.zhaw.pm2.multichat.protocol.ConnectionHandler;
import ch.zhaw.pm2.multichat.protocol.NetworkHandler;

import java.io.IOException;
import java.net.SocketException;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

/**
 * This Class represents a Server. The user can start the programm with the port number as a argument.
 * If no argument has been set the {@link NetworkHandler#DEFAULT_PORT} is used as port number.
 * After initialising the server:
 * 1. Starts a Socketserver using the logic given in {@link NetworkHandler.NetworkServer#createServer()}
 * 2. The server starts to listen for incoming connection using the Logic given in {@link NetworkHandler.NetworkServer#waitForConnection()}
 * 3. New conections will be atached to a Connectionhandler: {@link ServerConnectionHandler} and placed in a Map containing all active connections {@link Server#connections}
 *
 */
public class Server {

    // Server connection
    private NetworkHandler.NetworkServer<String> networkServer;

    // Connection registry
    private Map<String,ServerConnectionHandler> connections = new HashMap<>();

    public static void main(String[] args) {
        // Parse arguments for server port.
        try {
            int port;
            switch (args.length) {
                case 0 -> port = NetworkHandler.DEFAULT_PORT;
                case 1 -> port = Integer.parseInt(args[0]);
                default -> {
                    System.out.println("Illegal number of arguments:  [<ServerPort>]");
                    return;
                }
            }
            // Initialize server
            final Server server = new Server(port);

            // This adds a shutdown hook running a cleanup task if the JVM is terminated (kill -HUP, Ctrl-C,...)
            Runtime.getRuntime().addShutdownHook(new Thread() {
                @Override
                public void run() {
                    try {
                        Thread.sleep(200);
                        System.out.println("Shutdown initiated...");
                        server.terminate();
                    } catch (InterruptedException e) {
                        System.out.println("Warning: Shutdown interrupted. " + e);
                    } finally {
                        System.out.println("Shutdown complete.");
                    }
                }
            });

            // Start server
            server.start();
        } catch (IOException e) {
            System.err.println("Error while starting server." + e.getMessage());
        }
    }

    /**
     * The Constructor to create a new instance.
     * @param serverPort to listen for incoming connections.
     * @throws IOException thrown if an I/O error occurs when opening the socket.
     */
    public Server(int serverPort) throws IOException {
        // Open server connection
        System.out.println("Create server connection");
        networkServer = NetworkHandler.createServer(serverPort);
        System.out.println("Listening on " + networkServer.getHostAddress() + ":" + networkServer.getHostPort());
    }

    /**
     * With this methode the instance waits for incoming connections. If a client tries to connect to the server.
     * The connection will be registered in the connection registry if successful.
     */
    private void start() {
        ReentrantLock mutex = new ReentrantLock();
        Condition nameComplete = mutex.newCondition();
        System.out.println("Server started.");
        try {
            while (true) {
                 NetworkHandler.NetworkConnection<String> connection = networkServer.waitForConnection();
                 ServerConnectionHandler connectionHandler = new ServerConnectionHandler(connection, connections, mutex, nameComplete);
                 new Thread(connectionHandler).start();
            }
        } catch(SocketException e) {
            System.out.println("Server connection terminated");
        } catch (IOException e) {
            System.err.println("Communication error " + e);
        }
        // close server
        System.out.println("Server Stopped.");
    }

    /**
     * This method will stop the serversocket.
     */
    public void terminate() {
        try {
            System.out.println("Close server port.");
            networkServer.close();
        } catch (IOException e) {
            System.err.println("Failed to close server connection: " + e);
        }
    }

}
