package ch.zhaw.pm2.multichat.server;

import ch.zhaw.pm2.multichat.protocol.ChatProtocolException;
import ch.zhaw.pm2.multichat.protocol.ConnectionHandler;
import static ch.zhaw.pm2.multichat.protocol.ConnectionHandler.State.*;
import ch.zhaw.pm2.multichat.protocol.NetworkHandler;

import java.io.EOFException;
import java.io.IOException;
import java.net.SocketException;
import java.util.Map;
import java.util.Objects;
import java.util.Scanner;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;


/**
 * This class represents the connection between the server and a client and offers the serverside logic.
 * The ServerConnectionHandler receives data send from the client aswell as sends data to the client.
 *
 * The ServeConnectionHandler offers following functionality:
 *
 * Evaluating connection attempts from a client by:
 * 1. Checks if Username is valid (Not used)
 * 2. Saves Username in {@link ServerConnectionHandler#userName}
 * 3. Saves the connection in the {@link ServerConnectionHandler#connectionRegistry}
 *
 * Processes disconnections from a client by:
 * 1. Removing the connection from the {@link ServerConnectionHandler#connectionRegistry}
 * 2. Terminates the socket by calling {@link NetworkHandler.NetworkConnection#close()}
 *
 * Processes Messages send from a client by:
 * 1. Evaluating the reciever by differentiating between broadcast or unicast.
 * 2. Sending the message accordingly.
 *
 * To use this class, start a new instance and start it in a thread.
 * To constructor needs following parameter:
 * 1. {@link ch.zhaw.pm2.multichat.protocol.NetworkHandler.NetworkConnection} representing the socket connection between server and client
 * 2. {@link Map<String,ServerConnectionHandler>} registry to check for all active connections
 * 3. {@link ReentrantLock @link Condition} to lock server thread to evaluate connection.
 * */
public class ServerConnectionHandler extends ConnectionHandler implements Runnable{
    private static final AtomicInteger connectionCounter = new AtomicInteger(0);
    private final int connectionId = connectionCounter.incrementAndGet();
    private final Map<String,ServerConnectionHandler> connectionRegistry;

    private ReentrantLock mutex;

    private Condition nameComplete;

    private String userName = "Anonymous-"+connectionId;
    private State state = NEW;

    /**
     * Called when runnable gets started in a thread.
     */
    @Override
    public void run() {
        startReceiving();
    }

    /**
     * Constructor to intitialize the connection
     * @param connection representing the socket connection between server and clinet
     * @param registry map containing all active connections between server and clients
     * @param mutex to lock thread
     * @param nameComplete condition to call threads
     */
    public ServerConnectionHandler(NetworkHandler.NetworkConnection<String> connection,
                                   Map<String,ServerConnectionHandler> registry, ReentrantLock mutex, Condition nameComplete) {
        super();
        setConnection(connection);
        Objects.requireNonNull(connection, "Connection must not be null");
        Objects.requireNonNull(registry, "Registry must not be null");
        this.connectionRegistry = registry;
        this.mutex = mutex;
        this.nameComplete = nameComplete;
    }

    /**
     *
     * @return the username of the connected client
     */
    public String getUserName() {
        return this.userName;
    }

    /**
     *
     * @return state of the connection. Poosible states are see {@link ch.zhaw.pm2.multichat.protocol.ConnectionHandler.State}
     */
    public State getState() {
        return state;
    }

    /**
     * This methods runs in a whileloop as long as the socket between server and client is available
     * and the connection State is not ERROR.
     */
    private void startReceiving() {
            System.out.println("Starting Connection Handler for new User");
            try {
                System.out.println("Start receiving data...");
                while (getConnection().isAvailable() && !(state == ERROR)) {
                    String data = getConnection().receive();
                    processData(data);
                }
                System.out.println("Stopped recieving data");
            } catch (SocketException e) {
                System.out.println("Connection terminated locally");
                connectionRegistry.remove(userName);
                System.out.println("Unregistered because client connection terminated: " + userName + " " + e.getMessage());
            } catch (EOFException e) {
                System.out.println("Connection terminated by remote");
                connectionRegistry.remove(userName);
                System.out.println("Unregistered because client connection terminated: " + userName + " " + e.getMessage());
            } catch (IOException e) {
                System.err.println("Communication error: " + e);
            } catch (ClassNotFoundException e) {
                System.err.println("Received object of unknown type: " + e.getMessage());
            }
        if (state == ERROR) {
            System.out.println("Stopping Connection Handler for Rejected Client");
        } else {
            System.out.println("Stopping Connection Handler for " + userName);
        }

    }

    /**
     * This method will call {@link NetworkHandler.NetworkConnection#close()} to close the Socket.
     */
    private void stopReceiving() {
        System.out.println("Closing Connection Handler for " + userName);
        try {
            System.out.println("Stop receiving data...");
            getConnection().close();
            System.out.println("Stopped receiving data.");
        } catch (IOException e) {
            System.err.println("Failed to close connection." + e.getMessage());
        }
        System.out.println("Closed Connection Handler for " + userName);
    }

    /**
     * This method gets called when socket recieves data. The method checks for the data type and reacts accordingly
     * If data type:
     * 1. Connect => checks if username is valid. if valid sends response to client with confirmation.
     *  If username not valid quits connection by changing status to ERROR.
     * 2. Confirm => Server should not recieve this kind of message. STDOUT informs about it.
     * 3. Disconnect => Disconnects connection by removing connection from registry and calling method to terminate socket.
     * 4. Message => Checks if broadcast or unicast. Sends data accordingly
     * 5. Error => STDERR message
     * @param data recieved by the server
     */
    private void processData(String data)  {
        try {
            Scanner scanner = new Scanner(data);
            StringBuilder sender = new StringBuilder();
            StringBuilder reciever = new StringBuilder();
            StringBuilder type = new StringBuilder();
            StringBuilder payload = new StringBuilder();
            super.processData(scanner,sender,reciever,type,payload);

            // dispatch operation based on type parameter
            if (type.toString().equals(getDataTypeConnect())) {
                caseConnect(sender.toString());
            } else if (type.toString().equals(getDataTypeConfirm())) {
                System.out.println("Not expecting to receive a CONFIRM request from client");
            } else if (type.toString().equals(getDataTypeDisconnect())) {
                caseDisconnect();
            } else if (type.toString().equals(getDataTypeMessage())) {
                caseMessage(sender.toString(), reciever.toString(), type.toString(), payload.toString());
            } else if (type.toString().equals(getDataTypeError())) {
                System.err.println("Received error from client (" + sender + "): " + payload);
            } else {
                System.err.println("Unknown data type received: " + type);

            }
        } catch(ChatProtocolException e) {
            System.out.println("Error while processing data " + e.getMessage());
            sendData(USER_NONE, userName, getDataTypeError(), e.getMessage());
        }
    }

    /**
     * This method is called by method {@link ServerConnectionHandler#processData(String)}
     * Checks if username is valid. if valid sends response to client with confirmation.
     * @param sender of the payload
     * @throws ChatProtocolException if username not valid
     */
    private void caseConnect(String sender) throws ChatProtocolException {
        if (this.state != NEW) throw new ChatProtocolException("Illegal state for connect request: " + state);
        if (sender.isBlank()) sender = this.userName;
        //if username not valid
        if (connectionRegistry.containsKey(sender)) {
            state = ERROR;
            System.out.println(String.format("Connecting failed for new Client with IP:Port <%s:%d>.\nReason: Name already taken.",
                    getConnection().getRemoteHost(),
                    getConnection().getRemotePort()));
            throw new ChatProtocolException("User name already taken: " + sender);
        }
        //if username valid
        this.userName = sender;
        connectionRegistry.put(userName, this);
        sendData(USER_NONE, userName, getDataTypeConfirm(), "Registration successfull for " + userName);
        this.state = CONNECTED;
        System.out.println(String.format("Connected new Client %s with IP:Port <%s:%d>",
                userName,
                getConnection().getRemoteHost(),
                getConnection().getRemotePort()));
    }

    /**
     * This method is called by method {@link ServerConnectionHandler#processData(String)}
     * Disconnects connection by removing connection from registry and calling method {@link ServerConnectionHandler#stopReceiving()} to terminate socket.
     * @throws ChatProtocolException if state allready DISCONNECTED.
     */
    private void caseDisconnect() throws ChatProtocolException {
        if (state == DISCONNECTED)
            throw new ChatProtocolException("Illegal state for disconnect request: " + state);
        if (state == CONNECTED) {
            connectionRegistry.remove(this.userName);
        }
        sendData(USER_NONE, userName, getDataTypeConfirm(), "Confirm disconnect of " + userName);
        this.state = DISCONNECTED;
        this.stopReceiving();
    }

    /**
     * This method is called by method {@link ServerConnectionHandler#processData(String)}
     * Checks if broadcast or unicast. Sends data accordingly
     * @param sender who sent data
     * @param reciever to recieve data
     * @param type of message
     * @param payload data to transmit
     * @throws ChatProtocolException if state not equal to CONNECT
     */
    private void caseMessage(String sender, String reciever, String type, String payload) throws ChatProtocolException{
        if (state != CONNECTED) throw new ChatProtocolException("Illegal state for message request: " + state);
        if (USER_ALL.equals(reciever)) {
            for (ServerConnectionHandler handler : connectionRegistry.values()) {
                handler.sendData(sender, reciever, type, payload);
            }
        } else {
            ServerConnectionHandler handler = connectionRegistry.get(reciever);
            if (handler != null) {
                handler.sendData(sender, reciever, type, payload);
                if(!reciever.equals(sender)){
                    sendData(sender, reciever, type, payload); //send message to sender if it's a direct message and sender is not receiver.
                }
            } else {
                this.sendData(USER_NONE, userName, getDataTypeError(), "Unknown User: " + reciever);
            }
        }
    }
}
